'use strict';

var chores = require('./chores');
var errors = require('./errors');
var Validator = require('./validator');
var debugx = require('./pinbug')('injektor:engine');
var util = require('util');

var noop = function noop() {};
var ACCEPTED_SEPARATORS = ['/', ':', '$'];
var NAME_PATTERN_TMPL = '^[a-zA-Z]{1}[a-zA-Z0-9&\\-_%s]*$';

var defaultConfig = {
  argumentSchemaLabel: 'argumentSchema',
  argumentFieldsLabel: 'argumentProperties',
  referenceArrayLabel: 'referenceArray',
  namePatternTemplate: NAME_PATTERN_TMPL,
  detectCyclicDependencies: true,
  nameResolvingStrictMode: false,
  allowLookupDuplicateNames: false,
  separator: '/',
  logger: { debug: noop, trace: noop, error: noop }
};

function Injektor(a) {
  debugx.enabled && debugx(' + constructor start ...');

  a = a || {};

  var p = {};
  Object.keys(defaultConfig).forEach(function (b) {
    switch (b) {
      case 'separator':
        if (ACCEPTED_SEPARATORS.indexOf(a[b]) >= 0) {
          p[b] = a[b];
        } else {
          p[b] = defaultConfig[b];
        }
        break;
      default:
        if (b in a) {
          p[b] = a[b];
        } else {
          p[b] = defaultConfig[b];
        }
    }
  });

  var q = util.format(p.namePatternTemplate, p.separator);
  debugx.enabled && debugx(' - namePattern: %s', q);

  var r = new RegExp(q, 'g');

  var s = new Validator();

  var t = {};
  var u = {};

  /**
   * Get the dependency. If the call is the first time, copy it (for object) or
   * instantiate if (for constructor).
   */
  function b(a, b, d) {
    b = b || {};
    d = d || [];
    var e = n(a, b, d);
    if (typeof e === 'string') {
      b = k(e, b);
    }
    var f = t[e];
    if (f == null) {
      var g = new errors.DependencyNotFoundError('No dependency registered with name: ' + a);
      d.push(g);
      return undefined;
    }
    p.detectCyclicDependencies && l(e, b);
    f['cache'] = f['cache'] || c(e, b, d);
    p.detectCyclicDependencies && m(e, b);
    return f['cache'];
  }

  /**
   * Converts the dependency from definition into an instance (object or service).
   */
  function c(a, b, c) {
    var e = undefined;
    var f = t[a];
    if (f['type'] == 'service') {
      if (f['error'] == null) {
        try {
          var g = d(f['value'], b, c);
          if (c.length === 0) e = g;
        } catch (b) {
          debugx.enabled && debugx('convert("%s") has failed: %s/%s', a, b.name, b.message);
          f['error'] = b;
        }
      }
      if (f['error']) {
        var h = f['error'];
        c.push(h);
        if (o(h)) throw h;
      }
    } else {
      e = f['value'];
    }
    return e;
  }

  /**
   * Instantiates a constructor
   */
  function d(a, b, c) {
    var d = function wrapper(b) {
      return a.apply(this, b);
    };
    d.prototype = a.prototype;
    return new d(e(a, b, c));
  }

  /**
   * Extracts the dependencies of a constructor
   */
  function e(a, c, d) {
    if (a[p.argumentSchemaLabel]) {
      var e = a[p.argumentSchemaLabel];
      if (e instanceof Object && !(e instanceof Array) && (e.properties == null || e.properties instanceof Object) && (e.properties == null || !(e.properties instanceof Array))) {
        var g = {};
        var h = Object.keys(e.properties || {});
        for (var j = 0; j < h.length; j++) {
          g[h[j]] = b(h[j], c, d);
        }
        var i = s.validate(g, e);
        if (!i.ok) {
          throw new errors.ValidatingArgumentError('Constructor argument validation is failed');
        }
        return [g];
      } else {
        throw new errors.InvalidArgumentSchemaError('Constructor has invalid argumentSchema descriptor');
      }
    } else if (a[p.argumentFieldsLabel]) {
      var k = a[p.argumentFieldsLabel];
      if (k instanceof Array) {
        var g = {};
        for (var j = 0; j < k.length; j++) {
          g[k[j]] = b(k[j], c, d);
        }
        return [g];
      } else {
        throw new errors.InvalidArgumentSchemaError('Constructor has invalid argumentFields descriptor');
      }
    } else if (a[p.referenceArrayLabel]) {
      var l = a[p.referenceArrayLabel];
      if (l instanceof Array) {
        var m = [];
        for (var j = 0; j < l.length; j++) {
          m.push(b(l[j], c, d));
        }
        return m;
      } else {
        throw new errors.InvalidArgumentSchemaError('Constructor has invalid referenceArray descriptor');
      }
    } else {
      var n = f(a);
      return n.map(function (a) {
        return b(a, c, d);
      });
    }
  }

  function f(a) {
    var b = a.toString().replace(/((\/\/.*$)|(\/\*[\s\S]*?\*\/)|(\s))/mg, '').match(/^function\s*[^\(]*\(\s*([^\)]*)\)/m)[1].split(/,/);
    if (b.length == 1 && b[0] == "") return [];
    return b;
  }

  function g(a) {
    if (typeof a === 'function') {
      return { serviceInit: a };
    }
    if (a instanceof Array) {
      if (a.length > 0) {
        a = a.slice();
        var b = a.pop();
        if (typeof b === 'function') {
          var c = { serviceInit: b };
          if (a.length > 0) {
            c.referenceNames = a;
          }
          return c;
        }
      }
    }
    return {};
  }

  function h(a) {
    if (typeof a !== 'string' || a.length === 0) {
      throw new errors.InvalidMethodArgumentError('dependency name is empty or not a string');
    }
    if (a.match(r) == null) {
      throw new errors.InvalidMethodArgumentError('dependency name does not match the pattern');
    }
  }

  function i(a, b) {
    b = b || {};
    var c = a.split(p.separator);
    if (c.length === 1) {
      if (b.scope) c.unshift(b.scope);
    }
    if (c.length === 0 || c.length > 2) {
      throw new errors.InvalidMethodArgumentError('invalid name format');
    }
    var d = {
      name: a,
      absoluteName: c.join(p.separator),
      relativeName: c[c.length - 1],
      scope: c[c.length - 2]
    };
    return d;
  }

  function j(a) {
    if (a.absoluteName != a.relativeName) {
      u[a.relativeName] = u[a.relativeName] || [];
      if (u[a.relativeName].indexOf(a.absoluteName) < 0) {
        u[a.relativeName].push(a.absoluteName);
      }
    }
  }

  function k(a, b) {
    b = b || {};
    var c = chores.cloneObject(b);
    var d = a.split(p.separator);
    if (d.length === 2) c.scope = d[0];
    return c;
  }

  function l(a, b) {
    b = b || {};
    b.stacktrace = b.stacktrace || [];
    b.stacktrace.push(a);
    if (b.stacktrace.length == 1) {
      debugx.enabled && debugx(' - (%s) => %s', a, JSON.stringify(b.stacktrace));
    } else {
      debugx.enabled && debugx(' - (%s) -> %s', a, JSON.stringify(b.stacktrace));
    }
    var c = b.stacktrace.indexOf(a);
    if (c != b.stacktrace.length - 1) {
      throw new errors.DependencyCycleDetectedError(util.format('dependency cycle detected for "%s" at %s and %s in %s', a, b.stacktrace.length - 1, c, JSON.stringify(b.stacktrace)));
    }
  }

  function m(a, b) {
    b = b || {};
    b.stacktrace = b.stacktrace || [];
    b.stacktrace.pop();
    if (b.stacktrace.length > 0) {
      debugx.enabled && debugx(' - (%s) <- %s', a, JSON.stringify(b.stacktrace));
    } else {
      debugx.enabled && debugx(' - (%s) <= %s', a, JSON.stringify(b.stacktrace));
    }
  }

  function n(a, b, c) {
    b = b || {};
    var d = null;
    if (b.scope) {
      var e = [b.scope, a].join(p.separator);
      if (t[e]) d = e;
    }
    if (d == null && t[a]) d = a;
    if (d == null && u[a] && u[a].length > 0) {
      if (u[a].length === 1 || p.nameResolvingStrictMode !== true) {
        d = u[a][0];
      }
      if (p.allowLookupDuplicateNames != true && u[a].length > 1) {
        var f = new errors.DuplicatedRelativeNameError('name [' + a + '] is duplicated');
        if (c instanceof Array) {
          c.push(f);
        } else {
          throw f;
        }
      }
    }
    return d;
  }

  function o(a) {
    return a instanceof errors.DependencyNotFoundError || a instanceof errors.DependencyCycleDetectedError;
  }

  /**
   * Registering an object as a dependency
   */
  this.registerObject = function (a, b, c) {
    h(a);
    var d = i(a, c);
    t[d.absoluteName] = { type: 'object', value: b };
    j(d);
    return this;
  };

  /**
   * Registering a singleton-service constructor as a dependency
   */
  this.defineService = function (a, b, c) {
    h(a);
    var d = g(b);
    if (!d.serviceInit) {
      throw new errors.InvalidMethodArgumentError('invalid service descriptor');
    }
    if (d.referenceNames) {
      d.serviceInit.referenceArray = d.referenceNames;
    }
    var e = i(a, c);
    t[e.absoluteName] = { type: 'service', value: d.serviceInit };
    j(e);
    return this;
  };

  this.parseName = function (a, b) {
    var c = Object.assign({}, b);
    var d = c.exceptions;
    delete c.exceptions;
    return i(a, c, d);
  };

  this.resolve = function (a, b) {
    var c = Object.assign({}, b);
    var d = c.exceptions;
    delete c.exceptions;
    return n(a, c, d);
  };

  this.resolveName = this.resolve;

  this.suggest = function (a) {
    if (t[a]) return [a];
    if (u[a]) {
      if (u[a].length === 0) {
        return [];
      } else {
        return u[a].slice();
      }
    }
    return null;
  };

  this.suggestName = this.suggest;

  /**
   * Lookups a dependency (object or service) from store 
   */
  this.lookup = function (a, c) {
    h(a);
    var d = false;
    var e = [];
    // extract exceptions from options
    if (chores.isArray(c)) {
      d = true;
      e = c;
      c = {};
    } else if (chores.isObject(c)) {
      d = chores.isArray(c.exceptions);
      e = c.exceptions = c.exceptions || [];
    }
    // retrieve object from storage
    var f = b(a, c, e);
    if (!d && e.length > 0) {
      chores.printExceptions(e);
      if (e.length === 1) throw e[0];
      throw new errors.DependencyCombinationError('lookup() is failed', e);
    }
    return f;
  };

  /**
   * Collects all dependencies of a service, pass them to and run the service
   */
  this.invoke = function (a) {
    if (!chores.isArray(a)) {
      a = Array.prototype.slice.call(arguments);
    }
    var b = g(a);
    if (!b.serviceInit) {
      throw new errors.InvalidMethodArgumentError('invalid callback descriptor');
    }
    b.serviceInit.referenceArray = b.referenceNames;
    var c = {};
    var d = [];
    var f = e(b.serviceInit, c, d);
    if (d.length > 0) {
      chores.printExceptions(d);
      if (d.length === 1) throw d[0];
      throw new errors.DependencyCombinationError('invoke() is failed', d);
    }
    return b.serviceInit.apply(null, f);
  };

  Object.defineProperty(this, 'separator', {
    get: function get() {
      return p.separator;
    },
    set: function set(a) {}
  });

  debugx.enabled && debugx(' - constructor end');
};

module.exports = Injektor;